// Persistence of Vision Ray Tracer Scene Include File
// File: treinbaan.inc
// Desc: 	This include file provides train tracks.
//		Straight tracks and curved tracks.
//		The dutch word 'treinbaan' means railroad layout
//		in english.
// Auth: Brickman Brick
//

/*
 * Brief overview of the functionality of this file:
 * 1. create objects (straight tracks, curved tracks etc.)
 * 2. lay the objects and calculate new variables
 * 3. calculate railroad layout
 *
 * Every variable and macro is preceeded with the abbreviation 'TB',
 * which stands for 'treinbaan'.
 */


/* 
 * include files
 *
 */

#include "arrays.inc"
#include "math.inc"
#include "transforms.inc"

#include "common_vars.inc"
#include "railroad_textures.inc"


/*
 * track objects
 *
 */

#local treinbaan_generic_straight_track = box { 
	<-1 * CVar_track_width / 2, CVar_track_bottom, 0>,
	<CVar_track_width / 2, CVar_track_top, CVar_straight_track_length>
	}

#local treinbaan_generic_straight_track_double_length = union {
	object {treinbaan_generic_straight_track}
	object {treinbaan_generic_straight_track
		translate <0, 0, CVar_straight_track_length> }
	}

#local treinbaan_straight_track = object {treinbaan_generic_straight_track
	//pigment {color blue .95 }
	texture {macro_test (blue .95) }
	}

#local treinbaan_generic_curved_track = difference {
	cylinder { <0, CVar_track_bottom, 0>,
		<0, CVar_track_top, 0>,
		CVar_curved_track_radius + (CVar_track_width / 2) }
	cylinder { <0, CVar_track_bottom - 1, 0>,
		<0, CVar_track_top + 1, 0>,
		CVar_curved_track_radius - (CVar_track_width / 2) }
	// box south of the x-axis
	box { <-1 * (CVar_curved_track_radius + CVar_track_width), CVar_track_bottom - 2, -1 * (CVar_curved_track_radius + CVar_track_width)>,
		<CVar_curved_track_radius + CVar_track_width, CVar_track_top + 2, 0> }
	// box north of the x-axis and rotated
	box { <-1 * (CVar_curved_track_radius + CVar_track_width + 1), CVar_track_bottom - 3, 0>
		<CVar_curved_track_radius + CVar_track_width + 1, CVar_track_top + 3, CVar_curved_track_radius + CVar_track_width + 1>
		rotate <0, CVar_curved_track_angle , 0>
		}
	}

#local treinbaan_right_curved_track = object {treinbaan_generic_curved_track 
	translate <CVar_curved_track_radius, 0, 0>
	//pigment {color red 1}
	texture {macro_test (red 1)}
	}

#local treinbaan_left_curved_track = object {treinbaan_generic_curved_track 
	rotate <0, 180 - CVar_curved_track_angle, 0> 
	translate <-1 * CVar_curved_track_radius, 0, 0>  
	//pigment {color green 1 }};
	texture {macro_test (green 1)}
	}

#local treinbaan_flat_crossing = union {
	object {treinbaan_generic_straight_track}
	object {treinbaan_generic_straight_track
		rotate <0, 90, 0> 
		translate <-1 * CVar_straight_track_length / 2, 0, CVar_straight_track_length / 2>}
	//pigment {color rgb <1, 1, 0> } 	// geel
	texture {macro_test (rgb <1, 1, 0>)}
	}

/*
 * left handed switches
 */

#local switch_diverge_left = union {
	object {treinbaan_generic_straight_track_double_length}
	object {treinbaan_generic_straight_track_double_length
		Shear_Trans (x, y, <-0.25, 0, 1>) }
	//pigment { color Cyan }
	texture {macro_test (Cyan)}
	}

#local switch_converge_right = union {
	object {treinbaan_generic_straight_track_double_length}
	object {treinbaan_generic_straight_track_double_length
		Shear_Trans (x, y, <-0.25, 0, 1>)
		translate <CVar_straight_track_length / 2, 0, 0> }
	//pigment { color Magenta }
	texture { basis_texture }
	}

/*
 * right handed switches
 */

#local switch_diverge_right = union {
	object {treinbaan_generic_straight_track_double_length}
	object {treinbaan_generic_straight_track_double_length
		Shear_Trans (x, y, <0.25, 0, 1>) }
	//pigment { color Cyan }
	texture {macro_test (Cyan)}
	}

#local switch_converge_left = union {
	object {treinbaan_generic_straight_track_double_length}
	object {treinbaan_generic_straight_track_double_length
		Shear_Trans (x, y, <0.25, 0, 1>)
		translate <-1 * CVar_straight_track_length / 2, 0, 0> }
	//pigment { color Magenta }
	texture {macro_test (Magenta)}
	}

//
// The second step is to create macros from the previously declared objects and variables.
// These macros place the tracks in the railroad layout.
//

#macro TB_Place_Track_Straight (TB_Position, TB_Direction)
	object {treinbaan_straight_track
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#declare TB_Position = TB_Position + 
		<CVar_straight_track_length * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * cos (radians (TB_Direction))>;
#end

#macro TB_Place_Track_Curved_Right (TB_Position, TB_Direction)
	object {treinbaan_right_curved_track
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#local curve_centre_vector = 
		<CVar_curved_track_radius * cos (radians (TB_Direction)),
		0,
		CVar_curved_track_radius * -1 * sin (radians (TB_Direction))>;
	#declare TB_Position = TB_Position + curve_centre_vector +
		vrotate (-1 * curve_centre_vector, <0, CVar_curved_track_angle, 0>);
	#declare TB_Direction = TB_Direction + CVar_curved_track_angle;
#end

#macro TB_Place_Track_Curved_Left (TB_Position, TB_Direction)
	object {treinbaan_left_curved_track
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#local circle_centre_vector = 
		<CVar_curved_track_radius * -1 * cos (radians (TB_Direction)),
		0,
		CVar_curved_track_radius * sin (radians (TB_Direction))>;
	#declare TB_Position = TB_Position + circle_centre_vector +
		vrotate (-1 * circle_centre_vector, <0, -1 * CVar_curved_track_angle, 0> );
	#declare TB_Direction = TB_Direction - CVar_curved_track_angle;
#end

#macro TB_Place_Track_Flat_Crossing (TB_Position, TB_Direction, place_object)
	#if (place_object)
		object {treinbaan_flat_crossing
			rotate <0, TB_Direction, 0>
			translate TB_Position
		}
	#end
	#declare TB_Position = TB_Position + 
		<CVar_straight_track_length * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * cos (radians (TB_Direction))>;
#end

/* 
 * There are two switches. The left and right handed. Each switch can be laid out 
 * in three different ways. This totals to 6 possibilities.
 */

#macro treinbaan_place_switch_diverge_left (TB_Position, TB_Direction)
	object {switch_diverge_left
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#declare TB_Position = TB_Position +
		<CVar_straight_track_length * 2 * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * 2 * cos (radians (TB_Direction))>;
#end

#macro treinbaan_place_switch_diverge_right (TB_Position, TB_Direction)
	object {switch_diverge_right
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#declare TB_Position = TB_Position +
		<CVar_straight_track_length * 2 * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * 2 * cos (radians (TB_Direction))>;
#end

#macro treinbaan_place_switch_converge_left (TB_Position, TB_Direction)
	object {switch_converge_left
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#declare TB_Position = TB_Position +
		<CVar_straight_track_length * 2 * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * 2 * cos (radians (TB_Direction))>;
#end

#macro treinbaan_place_switch_converge_left_converging_track (TB_Position, TB_Direction)
	object {switch_converge_left
		translate <CVar_track_width, 0, 0>
		rotate <0, TB_Direction, 0>
		translate TB_Position
	#declare TB_Position = TB_Position +
		<CVar_straight_track_length * 2 * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * 2 * cos (radians (TB_Direction))> +
		<CVar_track_width * cos (radians (TB_Direction)),
		0,
		CVar_track_width * -1 * sin (radians (TB_Direction))>;
	}
#end

#macro treinbaan_place_switch_converge_right (TB_Position, TB_Direction)
	object {switch_converge_right
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#declare TB_Position = TB_Position +
		<CVar_straight_track_length * 2 * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * 2 * cos (radians (TB_Direction))>;
#end

#macro treinbaan_place_switch_converge_right_converging_track (TB_Position, TB_Direction)
	object {switch_converge_right
		translate <-1 * CVar_track_width, 0, 0>
		rotate <0, TB_Direction, 0>
		translate TB_Position
	}
	#declare TB_Position = TB_Position +
		<CVar_straight_track_length * 2 * sin (radians (TB_Direction)),
		0,
		CVar_straight_track_length * 2 * cos (radians (TB_Direction))> +
		<CVar_track_width * -1 * cos (radians (TB_Direction)),
		0,
		CVar_track_width * sin (radians (TB_Direction))>;
#end


/*
 * macros to calculate the dimensions of the railroad layout 
 *
 */

#macro TB_Set_Initial_Extreme_Coordinates (position)
	#declare tb_x_min = position.x;
	#declare tb_x_max = position.x;
	#declare tb_z_min = position.z;
	#declare tb_z_max = position.z;
#end

#macro TB_Extreme_Coordinates (position)
	#declare tb_x_min = min (position.x, tb_x_min);
	#declare tb_x_max = max (position.x, tb_x_max);
	#declare tb_z_min = min (position.z, tb_z_min);
	#declare tb_z_max = max (position.z, tb_z_max);
#end


//
// Below is the code that reads the track layout data from a text file
// and calls the above macros to create the railroad layout.
// Additionally the length and the dimensions of the layout is calculated.
//


/*
 * set initial variables for calculating the layout
 */

#declare TB_Position = CVar_track_position;
#declare TB_Direction = CVar_track_direction;
#declare TB_Track_Length = 0;
#local TB_Num_Flat_Crossings = 0; 	//this value is decremented each time a flat-crossing has been placed
#local TB_Num_Switches = 0; 		//this value is incremented each time a switch has been placed
#local TB_Num_Diverging_Switches = 0;
#local TB_Diverging_Path_Counter = 0; 
TB_Set_Initial_Extreme_Coordinates (CVar_track_position)


/*
 * start calculating the layout
 */

#fopen TREINBAAN_RAILROAD_LAYOUT_FILE RAILROAD_LAYOUT_FILE read
	#while (defined (TREINBAAN_RAILROAD_LAYOUT_FILE))
		#read (TREINBAAN_RAILROAD_LAYOUT_FILE, treinbaan_track_id, switch_number)
		//#local treinbaan_track_id = int (treinbaan_track_id);
		//#debug concat ("waarde: ", str (treinbaan_track_id, 0, 10), "\n")
		#switch (treinbaan_track_id)
			/*
			#debug concat ("tt_id: ", str(treinbaan_track_id, 2, 0), 
				"\tdir: ", str(TB_Direction, 5, 1), "\tpositie: <", vstr(3, TB_Position, ", ", 0, 1), ">\n")
			*/
			#case (1)
				TB_Place_Track_Straight (TB_Position, TB_Direction)
				#declare TB_Track_Length = TB_Track_Length + CVar_straight_track_length;
			#break
			#case (2)
				TB_Place_Track_Curved_Right (TB_Position, TB_Direction)
				#declare TB_Track_Length = TB_Track_Length + (CVar_curved_track_radius * pi / 8);
				/*
				 * The circumference of a circle is 2 * pi * radius.
				 * A circle has 16 track segments and a half circle has 8 segments.
				 */
			#break
			#case (3)
				TB_Place_Track_Curved_Left (TB_Position, TB_Direction)
				#declare TB_Track_Length = TB_Track_Length + (CVar_curved_track_radius * pi / 8);
			#break
			#case (4)
				#if (TB_Num_Flat_Crossings > switch_number)
					TB_Place_Track_Flat_Crossing (TB_Position, TB_Direction, yes)
					#local TB_Num_Flat_Crossings = TB_Num_Flat_Crossings - 1;
				#else
					TB_Place_Track_Flat_Crossing (TB_Position, TB_Direction, no)
				#end
				#declare TB_Track_Length = TB_Track_Length + CVar_straight_track_length;
			#break
			#case (5)
				#if (TB_Num_Switches < switch_number)
					treinbaan_place_switch_diverge_left (TB_Position, TB_Direction)
					#local TB_Num_Switches = TB_Num_Switches + 1;

					//make sure the position of the diverging branch is remembered
					#if (TB_Num_Diverging_Switches = 0)
						#declare list_of_switched_positions = array[TB_Num_Diverging_Switches + 1];
						#declare list_of_switched_directions = array[TB_Num_Diverging_Switches + 1];
					#else
						Resize_Array (list_of_switched_positions, TB_Num_Diverging_Switches + 1)
						Resize_Array (list_of_switched_directions, TB_Num_Diverging_Switches + 1)
					#end

					#declare list_of_switched_positions[TB_Num_Diverging_Switches] = 
						TB_Position +
						<CVar_track_width * -1 * cos (radians (TB_Direction)),
						0,
						CVar_track_width * sin (radians (TB_Direction))>;
					#declare list_of_switched_directions[TB_Num_Diverging_Switches] = TB_Direction;
					#local TB_Num_Diverging_Switches = TB_Num_Diverging_Switches + 1;
				#end
				#declare TB_Track_Length = TB_Track_Length + (CVar_straight_track_length * 2);
			#break
			#case (6)
				#if (TB_Num_Switches < switch_number)
					treinbaan_place_switch_diverge_right (TB_Position, TB_Direction)
					#local TB_Num_Switches = TB_Num_Switches + 1;

					//make sure the position of the diverging branch is remembered
					#if (TB_Num_Diverging_Switches = 0)
						#declare list_of_switched_positions = array[TB_Num_Diverging_Switches + 1];
						#declare list_of_switched_directions = array[TB_Num_Diverging_Switches + 1];
					#else
						Resize_Array (list_of_switched_positions, TB_Num_Diverging_Switches + 1)
						Resize_Array (list_of_switched_directions, TB_Num_Diverging_Switches + 1)
					#end

					#declare list_of_switched_positions[TB_Num_Diverging_Switches] = 
						TB_Position +
						<CVar_track_width * cos (radians (TB_Direction)),
						0,
						CVar_track_width * -1 * sin (radians (TB_Direction))>;
					#declare list_of_switched_directions[TB_Num_Diverging_Switches] = TB_Direction;
					#local TB_Num_Diverging_Switches = TB_Num_Diverging_Switches + 1;
				#end
				#declare TB_Track_Length = TB_Track_Length + (CVar_straight_track_length * 2);
			#break
			#case (7)
				#if (TB_Num_Switches < switch_number)
					treinbaan_place_switch_converge_left (TB_Position, TB_Direction)
					#local TB_Num_Switches = TB_Num_Switches + 1;
				#end
				#declare TB_Track_Length = TB_Track_Length + (CVar_straight_track_length * 2);
			#break
			#case (8)
				#if (TB_Num_Switches < switch_number)
					treinbaan_place_switch_converge_left_converging_track (TB_Position, TB_Direction)
					#local TB_Num_Switches = TB_Num_Switches + 1;
				#end
				#declare TB_Track_Length = TB_Track_Length +
					sqrt (pow (int (CVar_track_width), 2) +
					pow (int (CVar_straight_track_length * 2), 2));
			#break
			#case (9)
				#if (TB_Num_Switches < switch_number)
					treinbaan_place_switch_converge_right (TB_Position, TB_Direction)
					#local TB_Num_Switches = TB_Num_Switches + 1;
				#end
				#declare TB_Track_Length = TB_Track_Length + (CVar_straight_track_length * 2);
			#break
			#case (10)
				#if(TB_Num_Switches < switch_number)
					treinbaan_place_switch_converge_right_converging_track (TB_Position, TB_Direction)
					#local TB_Num_Switches = TB_Num_Switches + 1;
				//#else
					//#debug "wissel (10) NIET geplaatst ...\n"
				#end
				#declare TB_Track_Length = TB_Track_Length +
					sqrt (pow (int (CVar_track_width), 2) +
					pow (int (CVar_straight_track_length * 2), 2));
			#break

			#case (0)
				//#debug concat ("We are at switch no. ", str (TB_Diverging_Path_Counter + 1, 0, 0), " of ", str (TB_Num_Diverging_Switches, 0, 0), " diverging switches.\n")
				//#debug concat ("It\'s direction is ", str (list_of_switched_directions[TB_Diverging_Path_Counter], 0, 0), ".\n")
				//#debug concat ("And it\'s position is ", vstr (3, list_of_switched_positions[TB_Diverging_Path_Counter], ",", 0, 1), ".\n")
				//#debug concat ("track_id: ", str(treinbaan_track_id, 0, 0), " richting: ", str(TB_Direction, 0, 2), "positie : <", vstr(3, TB_Position, ", ", 0, 1), ">\n")
				#if (switch_number = 1000)
					//#debug "end of file\n"
				#else
					#declare TB_Position = list_of_switched_positions[TB_Diverging_Path_Counter];
					#declare TB_Direction = list_of_switched_directions[TB_Diverging_Path_Counter];
					#local TB_Diverging_Path_Counter = TB_Diverging_Path_Counter + 1;
					#declare TB_Track_Length = TB_Track_Length + 
						sqrt (pow (int (CVar_track_width), 2) + 
						pow (int (CVar_straight_track_length * 2), 2));
				#end
			#break

			#else 
				#error concat ("***** treinbaan.inc: incorrect track number -", str (treinbaan_track_id, 0, 0), "- *****\n")
		#end
		TB_Extreme_Coordinates (TB_Position)
	#end

#fclose TREINBAAN_RAILROAD_LAYOUT_FILE

/*
#if (write_log_file)
	#write (LOGFILE, "Total length of railroad layout: ", TB_Track_Length, " POV-Ray units\n")
#end
*/
